from datetime import date
from itertools import islice

import boto3
import click


@click.group(help="Operations on the Datateer Data Lake")
def datalake():
    pass


@datalake.command(help="Remove files created before the specified date")
@click.option(
    "--bucket",
    "-b",
    required=True,
    help="The name of the bucket to remove the objects from",
)
@click.option(
    "--date",
    "-d",
    required=True,
    type=click.DateTime(formats=["%Y-%m-%d"]),
    help="Objects created before this date will be deleted",
)
@click.option(
    "--prefix",
    "-p",
    required=False,
    help="Optional. Will limit the object removal to only objects that have this key prefix",
)
@click.option("--dry-run/--no-dry-run", default=True)
def delete_before_date(bucket: str, date: date, prefix: str = "", dry_run: bool = True):
    """Deletes files from bucket before date

    Parameters
    ----------
    bucket: str
        The name of the bucket to remove the objects from
    date: date
        Objects created before this date will be deleted
    prefix: str
        Optional. Will limit the object removal to only objects that have this key prefix
    """
    session = boto3.session.Session()
    bucket_name = bucket
    bucket = session.resource("s3").Bucket(bucket_name)

    objects_to_delete = []
    for obj in bucket.objects.all():
        if (obj.last_modified).replace(tzinfo=None) < date and obj.key.startswith(
            prefix
        ):
            objects_to_delete.append(obj.key)
            if dry_run:
                print(f"Dry run: would delete {obj.key}")
            else:
                print(f"Preparing to delete object {obj.key}")
    if dry_run:
        print(f"Dry run: would have deleted these {len(objects_to_delete)} objects")
    else:
        click.confirm(
            f"Are you SURE you want to delete these {len(objects_to_delete)}?",
            abort=True,
        )
        maximum_to_delete_in_batch = 1000  # AWS API can only delete 1,000 at a time
        while 0 < len(objects_to_delete):
            # filter to an acceptable batch size
            filtered = (o for o in objects_to_delete)
            slice_to_delete = list(islice(filtered, maximum_to_delete_in_batch))

            # do the delete
            bucket.delete_objects(
                Delete={"Objects": [{"Key": k} for k in slice_to_delete]}
            )

            # remove the deleted items
            objects_to_delete = [
                o for o in objects_to_delete if o not in slice_to_delete
            ]
            print(
                f"Deleted {len(slice_to_delete)} objects. {len(objects_to_delete)} remaining"
            )



import click


def info(msg: str) -> None:
    click.echo(msg)


def error(msg: str) -> None:
    click.echo(click.style(msg, fg="red"))


def success(msg: str) -> None:
    click.echo(click.style(msg, fg="green"))


def warn(msg: str) -> None:
    click.echo(click.style(msg, fg="yellow"))



import click

from . import __version__
from .commands.config.commands import config
from .commands.datalake.commands import datalake
from .commands.docs.commands import docs
from .commands.echo.echo_commands import echo
from .commands.infra.commands import infra
from .commands.pipeline.commands import pipeline

name = "Datateer CLI"


@click.group()
@click.pass_context
@click.version_option(
    version=__version__
)  # note this version param should not be necessary according to https://click.palletsprojects.com/en/8.0.x/api/#click.version_option
def main(ctx):
    pass


main.add_command(config)
main.add_command(docs)
main.add_command(echo)
main.add_command(infra)
main.add_command(pipeline)
main.add_command(datalake)



import os
import sys
from contextlib import contextmanager
from importlib import find_loader, import_module

import click

from .deploy_flow import do_deploy


@click.group(help="Commands related to data pipelines and Prefect flows")
def pipeline():
    pass


@pipeline.command(
    help="Packages and deploys a pipeline to Prefect Cloud. The NAME argument is the name of the pipeline. Will look for './orchestration/pipeline_<name>.py"
)
@click.argument("name", required=True)
@click.option(
    "--cloud",
    envvar="CLOUD",
    type=click.Choice(["aws", "gcp"], case_sensitive=True),
    required=True,
    help="The cloud provider which the pipeline lives",
)
@click.option(
    "-a",
    "--account",
    envvar="ACCOUNT",
    help="The account of the deployment target (defaults to env var ACCOUNT)",
)
@click.option(
    "-r",
    "--region",
    envvar="REGION",
    help="The region of the deployment target(defaults to env var REGION)",
)
@click.option(
    "-c",
    "--client-code",
    envvar="CLIENT_CODE",
    help="The client's Datateer client code (defaults to env var CLIENT_CODE)",
)
@click.option(
    "-d",
    "--deploy-key-prefect-lib",
    envvar="DEPLOY_KEY_PREFECT_LIB",
    help="The GitHub deployment key for the datateer-prefect repository (defaults to env var DEPLOY_KEY_PREFECT_LIB)",
)
@click.option(
    "-e",
    "--environment",
    envvar="DATATEER_ENV",
    type=click.Choice(["prod", "stg", "qa", "int"], case_sensitive=False),
    required=True,
    help="The target Datateer environment (defaults to env var DATATEER_ENV)",
)
@click.option(
    "-p",
    "--prefect-cloud-api-key",
    envvar="PREFECT__CLOUD__API_KEY",
    help="API key to connect to Prefect Cloud API (defaults to env var PREFECT__CLOUD__API_KEY)",
)
def deploy(
    name,
    cloud,
    account,
    region,
    client_code,
    deploy_key_prefect_lib,
    environment,
    prefect_cloud_api_key,
):
    with modified_environ(
        DATATEER_ENV=environment,
        CLIENT_CODE=client_code,
        DEPLOY_KEY_PREFECT_LIB=deploy_key_prefect_lib,
        PREFECT__CLOUD__API_KEY=prefect_cloud_api_key,
        CLOUD=cloud,  # referenced by datateer-prefect package
        GOOGLE_PROJECT_ID=account,  # referenced by datateer-prefect package
    ):
        pipeline = import_pipeline(name=name)
        pipeline.name = f"{client_code}-{pipeline.name.replace('_', '-')}-{environment}"
        custom_modules = list(get_custom_modules())
        click.echo(
            f"Deploying flow {pipeline.name} to the {environment} environment on {region}"
        )
        do_deploy(
            pipeline,
            cloud,
            account,
            region,
            client_code,
            environment,
            custom_modules,
        )


@pipeline.command(
    help="Visualizes a pipeline's DAG of tasks. The NAME argument is the name of the pipeline. Will look for './orchestration/pipeline_<name>.py"
)
@click.argument("name", required=True)
def visualize(name):
    pipeline = import_pipeline(pipelines_dir="./orchestration", name=name)
    pipeline.visualize()


# @pipeline.command()
# @click.argument("name", required=True, help="The name of the pipeline. Will look for './orchestration/pipeline_<name>.py")
# def run(name):
#     # if aws_profile:
#     #     os.environ["AWS_PROFILE"] = aws_profile
#     try:
#         pipeline = import_pipeline(pipelines_dir='./orchestration', name=name)
#         pipeline.run(run_on_schedule=False)
#     except Exception as ex:
#         if "Secrets Manager can't find the specified secret" in str(ex):
#             log.error(
#                 f'Could not find the settings for environment {os.getenv("DATATEER_ENV")}. Have you run datateer push-config {os.getenv("DATATEER_ENV")}?'
#             )
#         else:
#             raise


def get_custom_modules(pipelines_dir="orchestration"):
    """Generator that returns all custom modules that can be added to a pipeline deployment
    See https://docs.prefect.io/api/latest/storage.html#docker

    Yields a tuple of full_path, relative_path
    """
    for root, dirs, files in os.walk(pipelines_dir):
        for filename in files:
            if (
                filename.endswith("__init__.py")
                or "pipeline_" in filename
                or not filename.endswith(".py")
            ):
                continue
            full_path = os.path.abspath(os.path.join(root, filename))
            rel_path = os.path.join(root, filename).replace(f"{pipelines_dir}/", "")
            yield full_path, rel_path


def import_pipeline(pipelines_dir="orchestration", name=None):
    name = f"pipeline_{name}" if name is not None else "pipeline"
    path = os.path.join(os.getcwd(), pipelines_dir)
    click.echo(f"Importing pipeline {name} from {path}")
    sys.path.insert(1, path)  # finds the pipeline here
    mod = import_module(name)
    return mod.flow

    # try:
    # TODO: remove this testing line
    # except ModuleNotFoundError as ex:
    #     print(f"ERROR: Could not find the module you requested: {ex}.")
    #     print("Are you in the pipeline project's root directory?")
    #     print('Have you activated the venv in the pipeline project? (Run "source venv/bin/activate")')


@contextmanager
def modified_environ(*remove, **update):
    """
    Temporarily updates the ``os.environ`` dictionary in-place.
    The ``os.environ`` dictionary is updated in-place so that the modification
    is sure to work in all situations.
    :param remove: Environment variables to remove.
    :param update: Dictionary of environment variables and values to add/update.
    """
    env = os.environ
    update = update or {}
    remove = remove or []

    # remove any None values
    for k, v in dict(update).items():
        if v is None:
            del update[k]

    # List of environment variables being updated or removed.
    stomped = (set(update.keys()) | set(remove)) & set(env.keys())
    # Environment variables and values to restore on exit.
    update_after = {k: env[k] for k in stomped}
    # Environment variables and values to remove on exit.
    remove_after = frozenset(k for k in update if k not in env)

    try:
        env.update(update)
        [env.pop(k, None) for k in remove]
        yield
    finally:
        env.update(update_after)
        [env.pop(k) for k in remove_after]



import os
import subprocess
import sys
import traceback
from time import time

import boto3
import yaml

import datateer_cli.logging as log


def generate_erd(searchdir: str) -> None:
    """
    Generates entity relationship diagrams for upload to the client's site

    Parameters
    ----------
    searchdir: str
        The directory to search for dbt model configuration files. Will recursively search this directory.
    """
    from datateer_cli.commands.docs.generate_erd import generate, searchdir

    generate(searchdir)


def build_dbt_docs_command(target: str) -> None:
    """
    Runs the dbt docs generate command to create the html that will be uploaded to the client's site.

    Parameters
    ----------
    target: str
        The target database environment to use to generate the dbt docs.
    """

    command = f"dbt docs generate --target {target}".split()
    try:
        subprocess.Popen(command, stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
        return command
    except Exception:
        print(sys.exc_info())


def build_aws_s3_sync_command(client_code: str) -> None:
    """
    Uploads newly created documentation to the client's S3 bucket and removes older files.

    Parameters
    ----------
    client_code: str
        The three-character client code.
    """
    command = f"aws s3 cp dbt/target/ s3://dbt.{client_code}.datateer.com --recursive --profile admin-{client_code}".split()
    try:
        subprocess.Popen(command, stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
        return command
    except Exception:
        print(sys.exc_info())


def invalidate_cloudfront_cache() -> None:
    """
    Invalidates the cloudfront cache.

    """
    name = "docs_site_cloudfront_distribution_id"
    client = boto3.client("ssm")
    res = client.get_parameter(Name=name)["Parameter"]["Value"]
    client = boto3.client("cloudfront")
    client.create_invalidation(
        DistributionId=res,
        InvalidationBatch={
            "Paths": {
                "Quantity": 1,
                "Items": ["/*"],
            },
            "CallerReference": str(time()).replace(".", ""),
        },
    )


def deploy_docs(searchdir: str, target: str, client_code: str) -> None:
    """
    Generates documentation and deploys it to the client's documentation site.

    Parameters
    ----------
    searchdir: str
        The directory to search for dbt model configuration files. Will recursively search this directory.
    target: str
        The target database environment to use to generate the dbt docs.
    client_code: str
        The three-character client code.
    """

    generate_erd(searchdir)
    build_dbt_docs_command(target)
    build_aws_s3_sync_command(client_code)
    invalidate_cloudfront_cache()



# import os

# from datateer.prefect import PipelineFlow
# from datateer.prefect.tasks import DbtTask, LoadSettingsTask

# CRON_SCHEDULE = "0 7 * * *"  # 7am UTC, 2am EST
# FLOW_NAME = f"example1"

# # if os.getenv("DATATEER_ENV") is None or os.getenv("DATATEER_ENV") == "local":
# #     # load in environment variables if we are running this flow from a development environment. Otherwise, the environment variables will be pre-loaded already
# #     from dotenv import load_dotenv

# #     load_dotenv()

# # load_settings = LoadSettingsTask(name="load settings")
# # dbt_setup = DbtTask(name="dbt setup", operations=["debug", "deps"])
# # dbt_seed = DbtTask(name="dbt seed", operations=["seed"])
# # dbt_run = DbtTask(name="dbt run", operations=["run"])
# # dbt_test = DbtTask(name="dbt test", operations=["test"])

# # env = os.getenv("DATATEER_ENV")
# with PipelineFlow(FLOW_NAME, cron=CRON_SCHEDULE if env == "prod" else None, notify_service_desk_on_error=True) as flow:
#     dbt_setup.set_upstream(load_settings)
#     dbt_seed.set_upstream(dbt_setup)
#     dbt_run.set_upstream(dbt_seed)
#     dbt_test.set_upstream(dbt_run)



# """

# """
# import os
# import subprocess
# import sys
# from importlib import find_loader, import_module
# from time import time
# from typing import Dict, List

# import boto3
# import click
# import yaml

# import datateer_cli.logging as log
# from datateer_cli.commands.pipeline import deploy_flow
# from datateer_cli.commands.docs import docs as docs_commands
# from datateer_cli.commands.pipeline import settings as pipeline_settings


# class DatateerContext:
#     def __init__(self):
#         self.config = None  # stores the config from .datateer/config.yml
#         self.flow = None  # stores the Flow between import and deploy


# @click.group()
# @click.pass_context
# def cli(ctx):
#     ctx.ensure_object(DatateerContext)
#     ctx.obj.config = pipeline_settings.load_config()
#     for k, v in ctx.obj.config.items():
#         os.environ[k.upper()] = str(v)


# @cli.command()
# @click.pass_context
# def tester(ctx):
#     print(ctx.obj.config["datateer_deploy_key_prefect_lib"])


# @cli.command(help="Displays a simple hello message")
# @click.option(
#     "-t",
#     "--text",
#     default="world",
#     help="Whatever you send in will get echoed back out to you",
# )
# def hello(text):
#     click.echo(f"hello1111, {text}")


# @cli.command(help="Display the version and exit")
# def version():
#     from datateer_cli import __version__

#     print(__version__)


# region agent


# @cli.group(help="For working with the Prefect agent that executes pipelines")
# def agent():
#     pass


# endregion

# region developer


# @cli.group(help="Commands to help the Datateer engineers with projects")
# def developer():
#     pass


# @developer.command(
#     help="Creates virtual environments; intended to be used at the repo root"
# )
# @click.pass_context
# @click.option(
#     "-v",
#     "--venv-root",
#     default="venv",
#     help="The name of the folder where you want your venvs",
# )
# @click.option(
#     "-r",
#     "--requirements-dir",
#     default="requirements",
#     help='The name of the directory that contains your requirements files. "requirements.txt" will be installed in the main venv at venv/ or the directory specified in --venv-root. "requirements-dev.txt" will also be installed at the venv root. All others named "requirements-<name>.txt will be installed in independent venvs at venv/<name>.',
# )
# def venvs(ctx, venv_root, requirements_dir):
#     from datateer_cli.developer_setup import setup

#     setup(venv_root, requirements_dir)


# @developer.command(
#     help="Initialize Terraform remote state. Should be run in the <client-code>-infrastructure repository root, because it modifies the main.tf file"
# )
# @click.option("-c", "--client-code", required=True)
# @click.option("-r", "--aws-region", default="us-east-1")
# @click.option("-p", "--aws-profile", required=True)
# def init_remote_state(aws_profile, client_code, aws_region):
#     os.environ["AWS_DEFAULT_REGION"] = aws_region
#     os.environ["CLIENT_CODE"] = client_code
#     from datateer_cli.terraform_util import configure_remote_state_backend

#     configure_remote_state_backend(aws_profile)
# endregion

# region docs


# @cli.group(help="For generating docs including erds and dbt docs")
# def docs():
#     pass


# @docs.command(help="Deploys erd-diagrams, intended to be used at the root")
# @click.pass_context
# def generate_erd(searchdir):

#     docs_commands.generate_erd(searchdir)


# @docs.command(
#     name="deploy",
#     help="Deploys docs including erd-diagrams and dbt docs, intended to be used at the root",
# )
# @click.pass_context
# @click.option(
#     "-m", "--dbt-models-dir", default="dbt/models", help="The dbt models directory."
# )
# @click.option("-t", "--target", default="prod")
# @click.option("-c", "--client-code", default=None)
# def deploy_docs(ctx, dbt_models_dir, target, client_code):
#     if client_code is None:
#         client_code = os.getenv("CLIENT_CODE")
#     if client_code is None:
#         raise click.ClickException("Client code not found.")
#     docs_commands.deploy_docs(dbt_models_dir, target, client_code)


# end region

# region pipeline
# def import_pipeline(pipelines_dir="orchestration", name=None):
#     try:
#         sys.path.insert(1, os.getcwd())  # finds the pipeline here
#         # todo: I would like to be able to avoid installing datateer into the project's requirements, and just run the global version. But if I do that,
#         # then it cannot import correctly within the venv
#         # sys.path.insert(1, os.path.abspath('venv/lib/python3.8/site-packages/')) # finds prefect here
#         # sys.path.insert(1, os.path.abspath('venv/src/datateer-prefect')) # finds datateer.prefect here
#         # print(sys.prefix)
#         # print('-----', sys.path)
#         path = f"pipeline_{name}" if name is not None else "pipeline"
#         mod = import_module(f"{pipelines_dir}.{path}")
#         return mod.flow
#     except ModuleNotFoundError as ex:
#         print(f"ERROR: Could not find the module you requested: {ex}.")
#         print("Are you in the pipeline project's root directory?")
#         print(
#             'Have you activated the venv in the pipeline project? (Run "source venv/bin/activate")'
#         )


# @cli.group(help="For working with pipelines")
# @click.pass_context
# @click.option(
#     "-n",
#     "--name",
#     default=None,
#     help='The name of the pipeline. e.g. pipeline_<name>.py. If not provided, "pipeline.py" will be used as the pipeline file name',
# )
# @click.option(
#     "-p",
#     "--pipelines-dir",
#     default="orchestration",
#     help='The name of the directory that contains your pipeline files. The default is "orchestration"',
# )
# @click.option(
#     "-e",
#     "--environment",
#     default="local",
#     help="The name of the environment. Loads the pipeline in context of that environment, including environment variables",
#     type=click.Choice(["local", "prod", "staging"], case_sensitive=False),
# )
# def pipeline(ctx, name, pipelines_dir, environment="local"):
#     ctx.ensure_object(DatateerContext)
#     os.environ["DATATEER_ENV"] = environment
#     pipeline = import_pipeline(pipelines_dir=pipelines_dir, name=name)
#     ctx.obj.flow = pipeline


# @pipeline.command(
#     help="push environment settings into secure storage so that it can be used at runtime."
# )
# @click.option(
#     "-c",
#     "--config-dir",
#     default=".datateer",
#     help="Relative path to the configuration directory. Defaults to .datateer",
# )
# @click.option(
#     "-s",
#     "--settings-file",
#     help="Relative path to a settings file containing environment variables. Defaults to .env",
# )
# def push_settings(config_dir, settings_file=None):
#     environment = os.getenv("DATATEER_ENV")
#     settings_file = pipeline_settings.determine_settings_file(
#         environment, settings_file
#     )
#     settings = pipeline_settings.load_settings(settings_file)
#     pipeline_settings.push_settings(settings, environment)
#     pipeline_settings.push_config(config_dir, environment)


# @pipeline.command()
# @click.option("-c", "--config-dir", default=".datateer")
# def pull_config(config_dir):
#     environment = os.environ["DATATEER_ENV"]
#     pipeline_settings.pull_config(config_dir, environment)


# @pipeline.command()
# @click.pass_context
# def debug(ctx):
#     from pathlib import Path

#     path = Path(__file__)
#     print("This is the debug function for testing things out")
#     print("parent: ", path.parent)
#     print(ctx.obj.flow)


# @pipeline.command(help="packages and deploys a pipeline")
# @click.option("--debug", is_flag=True)
# @click.pass_context
# def deploy(ctx, debug):
#     ctx.ensure_object(DatateerContext)
#     environment = os.environ["DATATEER_ENV"]
#     ctx.obj.flow.name += "-" + environment
#     click.echo(f"Deploying flow {ctx.obj.flow.name} to the {environment} environment")
#     deploy_flow.do_deploy(ctx.obj.flow, environment=environment, debug=debug)


# @pipeline.command()
# @click.pass_context
# def visualize(ctx):
#     ctx.ensure_object(DatateerContext)
#     ctx.obj.flow.visualize()


# endregion


# if __name__ == "__main__":
#     cli()



from datateer_cli.commands.pipeline import aws_deploy_flow, gcp_deploy_flow


def do_deploy(
    flow,
    cloud,
    account,
    region,
    client_code,
    environment,
    custom_modules=[],  # list of tuples: full_path, rel_path
    prefect_api_key=None,
    deploy_key_prefect_lib=None,
    prefect_results_bucket=None,
    debug=False,
):
    if cloud == "aws":
        aws_deploy_flow.do_deploy(
            flow,
            account,
            region,
            client_code,
            environment,
            custom_modules,
        )
    elif cloud == "gcp":
        gcp_deploy_flow.do_deploy(
            flow,
            account,
            region,
            client_code,
            environment,
            custom_modules,
        )
    else:
        raise Exception("Cloud Provider supplied has not been implemented yet")



import json
import os
from pathlib import Path
from typing import Dict

import boto3
import botocore
import click
import yaml
from botocore.exceptions import ClientError
from google.cloud import storage

# DATATEER_CONFIG_DIR = './datateer'
DEPLOY_CONFIG_FILE = "./.datateer/config.yml"
# DOCKERFILE = './datateer/Dockerfile'

DEFAULTS = {
    "DBT_EXECUTABLE": "/datateer/venv/dbt/bin/dbt",  # 'venv/dbt/bin/dbt',
    "DBT_PROFILES_DIR": "/datateer/.datateer/dbt/",  # '.datateer/dbt',
    "MELTANO_EXECUTABLE": "/datateer/venv/meltano/bin/meltano",  # 'venv/meltano/bin/meltano'
}

REQUIREDS = [
    "RAW_BUCKET",
    "PREPPED_BUCKET",
    "PREFECT_RESULTS_BUCKET",
    "SERVICE_DESK_CUSTOMER_ID",
]

DEFAULT_REGION = "us-east-1"

# flake8: noqa: E402
import datateer_cli.logging as log


def determine_settings_file(environment: str, settings_file: str = None) -> str:
    """Determines which settings file to use, and returns a path to it"""
    if settings_file:
        if not Path(settings_file).exists():
            raise ValueError(f"Could not find file {settings_file}")
        return settings_file

    path = f".datateer/.env.{environment}"
    if not Path(path).exists():
        raise ValueError(f"Could not find file {path}")
    return path


def load_settings(path) -> Dict[str, str]:
    with open(path) as f:
        lines = [
            line.lstrip("export ")
            for line in f.read().splitlines()
            if not line.startswith("#") and "=" in line
        ]
        settings = {
            key: val for key, val in (line.split("=", 1) for line in lines)
        }  # os.getenv(key)
    log.info(f"Loaded {len(settings)} settings from {path}")
    return settings


def apply_defaults(settings: dict):
    defaults = 0
    for key in DEFAULTS:
        if key not in settings:
            # log.info(f'Did not find setting "{key}", using default value "{DEFAULTS[key]}"')
            settings[key] = DEFAULTS[key]
            defaults += 1
    if defaults:
        log.info(f"Using {defaults} defaults that were not set explicitly")


def verify_requireds(settings: dict):
    missings = [key for key in REQUIREDS if key not in settings]
    if missings:
        log.error(f'Missing {len(missings)} required settings: {", ".join(missings)}')
        raise click.Abort()


def aws_push_config(config_dir: str, region, environment):
    """Pushed a configuration up into the S3 config bucket. Puts it into a top-level folder with the name of the environment"""
    session = boto3.session.Session(region_name=region)
    delete_config(environment, session)
    config_bucket = get_parameter("/pipeline/config_bucket", session=session)
    bucket = session.resource("s3").Bucket(config_bucket)

    file_count = 0
    for root, dirs, files in os.walk(config_dir):
        for file in files:
            dir = os.path.relpath(root, config_dir)
            if (
                dir == "."
                or file.lower().startswith("readme")
                or file.lower().endswith(".example")
                or file.lower().startswith(".env")
            ):
                continue
            if dir == ".":
                path = os.path.join(environment, file)
            else:
                path = os.path.join(environment, dir, file)
            # print(f'uploading file {path}')
            bucket.upload_file(os.path.join(root, file), path)
            file_count += 1
    log.success(f"Uploaded {file_count} files to bucket {config_bucket}/{environment}")


def gcp_push_config(config_dir: str, config_bucket, region: str = None):
    """TODO: region parameter may not be necessary"""
    gcs = storage.Client()
    bucket = gcs.bucket(config_bucket)
    file_count = 0
    for root, dirs, files in os.walk(config_dir):
        for file in files:
            dir = os.path.relpath(root, config_dir)
            if (
                file.lower().startswith("readme")
                or file.lower().endswith(".example")
                or file.lower().startswith(".env")
            ):
                continue
            path = file if dir == "." else os.path.join(dir, file)
            blob = bucket.blob(path)
            print(f"uploading from {os.path.join(root, file)} to {path}")
            blob.upload_from_filename(os.path.join(root, file))
            file_count += 1
    log.success(f"Uploaded {file_count} files to bucket {config_bucket}")


def aws_pull_config(
    config_dir: str, aws_region, environment, session=None, config_bucket=None
):
    """Pulls a configuration down from S3. Intended to be used by the CI pipeline, but could be used to pull the config to a local folder for insepction"""
    session = session or boto3.session.Session(region_name=aws_region)
    config_bucket = config_bucket or get_parameter(
        "/pipeline/config_bucket", session=session
    )
    bucket = session.resource("s3").Bucket(config_bucket)
    file_count = 0
    for obj in bucket.objects.filter(Prefix=environment):
        if obj.key.endswith("/"):
            continue
        path = os.path.join(config_dir, "/".join(obj.key.split("/")[1:]))
        Path(os.path.dirname(path)).mkdir(parents=True, exist_ok=True)
        bucket.Object(obj.key).download_file(path)
        file_count += 1
    log.success(
        f"Downloaded {file_count} files into {config_dir} from bucket {config_bucket}/{environment}"
    )


def gcp_pull_config(config_dir: str, config_bucket: str, region: str = None):
    """TODO: region may not be necessary"""
    gcs = storage.Client()
    bucket = gcs.bucket(config_bucket)
    file_count = 0
    for blob in gcs.list_blobs(config_bucket):
        destination_path = os.path.join(config_dir, blob.name.lstrip("./"))
        destination_dir = "/".join(destination_path.split("/")[:-1])
        Path(destination_dir).mkdir(parents=True, exist_ok=True)
        blob.download_to_filename(destination_path)
        file_count += 1
    log.success(
        f"Downloaded {file_count} files into {config_dir} from bucket {config_bucket}"
    )


def delete_config(environment: str, session=None):
    """Deletes an environment's configuration. Used by "push_config" to clear things out before uploading an updated configuration"""
    config_bucket = get_parameter("/pipeline/config_bucket", session)
    bucket = session.resource("s3").Bucket(config_bucket)

    bucket.objects.filter(Prefix=environment).delete()


# def push_settings(settings: dict, environment: str):
#     apply_defaults(settings)
#     verify_requireds(settings)
#     client = boto3.client("secretsmanager")
#     name = f"datateer-pipeline-settings-{environment}"
#     value = json.dumps(settings)
#     try:
#         client.create_secret(
#             Name=name,
#             Description="Managed by Datateer. A collection of settings intended to be injected into a data pipeline as environment variables.",
#             SecretString=value,
#         )
#     except client.exceptions.ResourceExistsException as ex:
#         pass
#     client.put_secret_value(SecretId=name, SecretString=value)
#     log.success(f"Wrote {len(settings)} settings to secret {name}")
#     # print(json.dumps(settings, indent=4, sort_keys=True))


# def load_config(path: str = None) -> Dict[str, str]:
#     path = path or DEPLOY_CONFIG_FILE
#     config = {}

#     if not os.path.exists(path):
#         log.warn(
#             f"Could not find configuration file at {path}; Will attempt to use environment variables"
#         )
#         # raise click.Abort()
#     else:
#         with open(path) as f:
#             config = yaml.load(f, Loader=yaml.FullLoader)

#         if "aws_default_region" not in config:
#             log.warn(
#                 f'Did not find "aws_default_region" in {path}. Defaulting to region "{DEFAULT_REGION}".'
#             )
#             config["aws_default_region"] = DEFAULT_REGION

#         expected_props = ["client_code", "aws_account_id"]
#         for prop in expected_props:
#             missing = 0
#             if prop not in config:
#                 log.error(f'Missing required configuration property "{prop}" in {path}')
#                 missing += 1
#             if missing:
#                 raise click.Abort()

#     return config


def get_parameter(name: str, session=None) -> str:
    if not session:
        session = boto3
    client = session.client("ssm")
    try:
        value = client.get_parameter(Name=name, WithDecryption=True)
    except ClientError:
        return None

    return value["Parameter"]["Value"]



import click

from .docs import deploy_docs


@click.group(help="Commands related to documentation")
def docs():
    pass


@docs.command(
    help="Deploys docs including erd-diagrams and dbt docs, intended to be used at the root of a pipeline repository"
)
@click.option(
    "-m",
    "--dbt-models-dir",
    default="dbt/models",
    help="The dbt models directory",
    show_default=True,
)
@click.option(
    "-t",
    "--target",
    envvar="DATATEER_ENV",
    type=click.Choice(["prod", "stg", "qa", "int"], case_sensitive=False),
    required=True,
    help="The Datateer environment to use to generate the docs (defaults to env var DATATEER_ENV)",
)
@click.option(
    "-c",
    "--client-code",
    envvar="CLIENT_CODE",
    required=True,
    help="(defaults to env var CLIENT_CODE)",
)
def deploy(client_code, dbt_models_dir, target):
    deploy_docs(dbt_models_dir, target, client_code)



import click


@click.group(help="Commands related to client infrastructure")
def infra():
    pass


@infra.command(
    help="Initialize Terraform remote state for the client's infrastructure. Should be run in the <client-code>-infrastructure repository root, because it modifies the main.tf file by injecting a 'backend S3' resource block. See more at https://www.terraform.io/language/settings/backends/s3"
)
@click.option(
    "-p",
    "--aws-profile",
    envvar="AWS_PROFILE",
    required=True,
    help="(defaults to env var AWS_PROFILE)",
)
@click.option(
    "-r",
    "--aws-region",
    envvar="AWS_REGION",
    required=True,
    help="(defaults to env var AWS_REGION)",
)
@click.option(
    "-c",
    "--client-code",
    envvar="CLIENT_CODE",
    required=True,
    help="(defaults to env var CLIENT_CODE)",
)
def init_remote_state(aws_profile, aws_region, client_code):
    from datateer_cli.commands.infra.terraform_util import (
        configure_remote_state_backend,
    )

    configure_remote_state_backend(aws_profile, aws_region, client_code)



import prefect

client = prefect.Client(
    api_token="eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiJ9.eyJ0ZW5hbnRfaWQiOiI5MjdjMDU1OC0zNzQ5LTRjZmItYTI4Yi1jN2E2NGFhMWVlODYiLCJ1c2VyX2lkIjpudWxsLCJhcGlfdG9rZW5fc2NvcGUiOiJSVU5ORVIiLCJyb2xlIjoiUlVOTkVSIiwiaWF0IjoxNTcxMDg0ODQ4LCJleHAiOjQxMDI0NDQ4MDAsImp0aSI6ImI3ZGM3ZDVmLTU4ZWUtNDUxMy1iMDIzLWZlZDg2ZmQxMGZlMSIsImlzcyI6IlByZWZlY3QgQ2xvdWQiLCJhdWQiOiJQcmVmZWN0IENsb3VkIEFQSSIsInR5cCI6IkFQSSJ9.jOxFPB7Sl5gWnZVe_kSYAlS2j7A37R2MbseGA9EV5F7ma8s519ieAbkv43OrpD3rx__Z5iPOxD1g5smCm9niHBubGEmB2H5briaZbmTdymLpZKDvZjIvhp58GKW9JWxlc5RGU75bB5bnUMmB6Nt5Y7Trg_jxarEBVINTe86IlTnatOgTsULQYUDYGhLrCjHzXqXomBzuYC27Jxlv3t124z5gt7nhn1IS9VIaAGtTt73pco5caOvwwtyswZe9ZUvOdZTqIV1CADbBF-d-OkRGOWtdLstOZSg4bv4ScoJL69Q-GTE9M1ZggnqNj0X4wv78rmlgrIFq4Vq0XyWo22_xog"
)
client.save_api_token()


def create_client_project(client_code):
    print(f"Creating project for {client_code}")
    try:
        project_id = client.create_project(project_name=f"datateer-{client_code}")
        print(f"Created project datateer-{client_code}, id = {project_id}")
    except Exception:
        print(f"Project datateer-{client_code} already exists, skipping")


if __name__ == "__main__":
    create_client_project("hmn")



import glob
import logging
import os
import tempfile
from importlib import find_loader, import_module
from pathlib import Path
from shutil import copy, copytree, ignore_patterns
from typing import Dict

import click
import pathspec
import yaml

from datateer_cli.commands.config import operations

DOCKERFILE_TEMPLATES_DIR = Path(__file__).parent.joinpath("flow")


def do_deploy(
    flow,
    gcp_project,
    gcp_region,
    client_code,
    environment,
    custom_modules=[],  # list of tuples: full_path, rel_path
    prefect_api_key=None,
    deploy_key_prefect_lib=None,
    prefect_results_bucket=None,
    debug=False,
):
    deploy_key_prefect_lib = deploy_key_prefect_lib or os.getenv(
        "DEPLOY_KEY_PREFECT_LIB"
    )
    if deploy_key_prefect_lib is None:
        raise Exception("deploy_key_prefect_lib is None but is required in do_deploy")

    prefect_api_key = prefect_api_key or os.getenv("PREFECT__CLOUD__API_KEY")
    if prefect_api_key is None:
        raise Exception(
            "prefect_api_key is required, either through a CLI parameter or an environment variable PREFECT__CLOUD__API_KEY"
        )

    if prefect_results_bucket is None:
        click.echo(
            "prefect_results_bucket in do_deploy is None, so the results of this flow's run will not be captured anywhere"
        )
    if os.getenv("MELTANO_DATABASE_URI") is None:
        print(
            "MELTANO_DATABASE_URI is None, so this flow will not attempt to connect to a meltano database"
        )
    else:
        print(
            f'MELTANO_DATABASE_URI is populated with value {os.getenv("MELTANO_DATABASE_URI")[0:12]}***'
        )

    # expensive loads, so defer until now
    from prefect.client.client import Client
    from prefect.executors import LocalExecutor

    # from prefect.environments import LocalEnvironment
    from prefect.run_configs import KubernetesRun
    from prefect.storage import Docker

    dockerfile_path = copy(os.path.join(DOCKERFILE_TEMPLATES_DIR, "Dockerfile"), ".")
    copy(os.path.join(DOCKERFILE_TEMPLATES_DIR, ".dockerignore"), ".")

    flow.storage = Docker(
        registry_url=image_registry_url(gcp_project, gcp_region),
        python_dependencies=["boto3"],
        dockerfile=dockerfile_path,
        image_name=flow.name,
        image_tag="latest",
        files={
            full_path: os.path.join("/datateer/modules", module_path)
            for full_path, module_path in custom_modules
        },
        env_vars={
            "DATATEER_ENV": environment,
            "PYTHONPATH": "$PYTHONPATH:/datateer/modules/",
            "REGION": gcp_region,
            "DBT_PROFILES_DIR": "/datateer/dbt",
            "CLOUD": "gcp",
        },
        build_kwargs={
            "buildargs": {
                # during the build process, the build script runs "meltano install." By default, this command runs against an internal Sqlite database
                # To make sure this runs against the production meltano database, we ensure this env var exists and contains the connection string
                "MELTANO_DATABASE_URI": os.getenv("MELTANO_DATABASE_URI"),
                "DEPLOY_KEY_PREFECT_LIB": deploy_key_prefect_lib,
            }
        },
    )

    env_vars = {"DATATEER_ENV": environment}
    if prefect_results_bucket is not None:
        env_vars["PREFECT_RESULTS_BUCKET"] = prefect_results_bucket

    flow.run_config = KubernetesRun(
        cpu_request=".5",
        cpu_limit="1",
        memory_request="2G",
        memory_limit="8G",
        service_account_name="prefect",
    )

    flow.executor = LocalExecutor()

    Client().create_project(client_code)
    flow.register(
        project_name=client_code,
        # add the environment label and the client label to restrict which agents will run this
        labels=[environment, client_code, "gcp"],
    )


def image_registry_url(gcp_project, gcp_region, include_https=False):
    """Builds a well formed docker registry URL
    Arguments:
        gcp_project {string} -- the GCP project id
        gcp_region {string} -- the Region region e.g. us-central1

    Keyword Arguments:
        include_https {bool} -- True to include the https:// prefix, false to leave off (default: {False})

    Returns:
        string -- A well formed docker registry URL
    """
    registry = f"{gcp_region}-docker.pkg.dev/{gcp_project}/prefect-flows/"
    return f"https://{registry}" if include_https else registry



import glob
import logging
import os
import tempfile
from importlib import find_loader, import_module
from pathlib import Path
from shutil import copy, copytree, ignore_patterns
from typing import Dict

import boto3
import click
import pathspec
import yaml

from datateer_cli.commands.config import operations

DOCKERFILE_TEMPLATES_DIR = Path(__file__).parent.joinpath("flow")


def do_deploy(
    flow,
    aws_account_id,
    aws_region,
    client_code,
    environment,
    custom_modules=[],  # list of tuples: full_path, rel_path
    prefect_api_key=None,
    deploy_key_prefect_lib=None,
    prefect_results_bucket=None,
    debug=False,
):
    deploy_key_prefect_lib = deploy_key_prefect_lib or os.getenv(
        "DEPLOY_KEY_PREFECT_LIB"
    )
    if deploy_key_prefect_lib is None:
        raise Exception("deploy_key_prefect_lib is None but is required in do_deploy")

    prefect_api_key = prefect_api_key or os.getenv("PREFECT__CLOUD__API_KEY")
    if prefect_api_key is None:
        raise Exception(
            "prefect_api_key is required, either through a CLI parameter or an environment variable PREFECT__CLOUD__API_KEY"
        )

    if prefect_results_bucket is None:
        click.echo(
            "prefect_results_bucket in do_deploy is None, so the results of this flow's run will not be captured anywhere"
        )
    if os.getenv("MELTANO_DATABASE_URI") is None:
        print(
            "MELTANO_DATABASE_URI is None, so this flow will not attempt to connect to a meltano database"
        )
    else:
        print(
            f'MELTANO_DATABASE_URI is populated with value {os.getenv("MELTANO_DATABASE_URI")[0:12]}***'
        )

    # expensive loads, so defer until now
    from prefect.client.client import Client
    from prefect.executors import LocalExecutor

    # from prefect.environments import LocalEnvironment
    from prefect.run_configs import ECSRun
    from prefect.storage import Docker

    dockerfile_path = copy(os.path.join(DOCKERFILE_TEMPLATES_DIR, "Dockerfile"), ".")
    copy(os.path.join(DOCKERFILE_TEMPLATES_DIR, ".dockerignore"), ".")

    flow.storage = Docker(
        registry_url=image_registry_url(aws_account_id, aws_region),
        python_dependencies=["boto3"],
        dockerfile=dockerfile_path,
        image_tag="latest",
        files={
            full_path: os.path.join("/datateer/modules", module_path)
            for full_path, module_path in custom_modules
        },
        env_vars={
            "DATATEER_ENV": environment,
            "PYTHONPATH": "$PYTHONPATH:/datateer/modules/",
            "CLOUD": "aws",
        },
        build_kwargs={
            "buildargs": {
                # during the build process, the build script runs "meltano install." By default, this command runs against an internal Sqlite database
                # To make sure this runs against the production meltano database, we ensure this env var exists and contains the connection string
                "MELTANO_DATABASE_URI": os.getenv("MELTANO_DATABASE_URI"),
                "DEPLOY_KEY_PREFECT_LIB": deploy_key_prefect_lib,
            }
        },
    )

    task_definition = {
        "networkMode": "awsvpc",
        "requiresCompatibilities": ["FARGATE"],
        "cpu": "4 vcpu",
        "memory": "8 GB",
        "executionRoleArn": f"arn:aws:iam::{aws_account_id}:role/ecsTaskExecutionRole",
        "taskRoleArn": f"arn:aws:iam::{aws_account_id}:role/ecsTaskRole",
        "containerDefinitions": [
            {
                "name": "flow"
                # "logConfiguration": { # this won't create the log group, so we can't reference it until we have some mechanism to create it
                #     "logDriver": "awslogs",
                #     "options": {
                #         "awslogs-group": f"/datateer/pipelines/{flow.name}",
                #         "awslogs-region": os.environ["AWS_DEFAULT_REGION"],
                #         "awslogs-stream-prefix": "prefect-agent",
                #     },
                # },
            }
        ],
    }
    env_vars = {"DATATEER_ENV": environment, "AWS_DEFAULT_REGION": aws_region}
    if prefect_results_bucket is not None:
        env_vars["PREFECT_RESULTS_BUCKET"] = prefect_results_bucket
    subnets = operations.get_parameter("/pipeline/allowed_subnets")

    if subnets:
        subnets = subnets.split(",")
    else:
        subnets = [
            operations.get_parameter("/pipeline/subnet_1"),
            operations.get_parameter("/pipeline/subnet_2"),
        ]

    flow.run_config = ECSRun(
        task_definition=task_definition,
        env=env_vars,
        run_task_kwargs={
            "networkConfiguration": {
                "awsvpcConfiguration": {
                    "subnets": subnets,
                    "securityGroups": [
                        operations.get_parameter("/pipeline/security_group")
                    ],
                    "assignPublicIp": "DISABLED",
                }
            }
        },
    )

    flow.executor = LocalExecutor()

    Client().create_project(client_code)
    register(flow, environment, client_code)


def register(flow, environment, client_code):
    import datateer_cli.docker_util as d

    d.docker_login()
    d.ensure_repository_exists(flow.name)
    flow.register(
        project_name=client_code,
        # add the environment label and the client label to restrict which agents will run this
        labels=[environment, client_code, "aws"],
    )


def image_registry_url(aws_account_id, aws_region, include_https=False):
    """Builds a well formed docker registry URL
    Arguments:
        aws_account_id {string} -- the AWS account ID
        aws_region {string} -- the AWS region e.g. us-east-1 or eu-west-1

    Keyword Arguments:
        include_https {bool} -- True to include the https:// prefix, false to leave off (default: {False})

    Returns:
        string -- A well formed docker registry URL
    """
    # aws_account_id = boto3.client("sts").get_caller_identity()["Account"]
    registry = f"{aws_account_id}.dkr.ecr.{aws_region}.amazonaws.com"
    return f"https://{registry}" if include_https else registry


# def validate_environment(ctx, name, environment):
#     path = '.env' if environment == 'local' else f'.env.{environment}'
#     if not os.path.exists(path):
#         raise click.BadParameter(f'Could not locate an .env file at {path}')
#     return environment



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['datateer_cli',
 'datateer_cli.commands.config',
 'datateer_cli.commands.datalake',
 'datateer_cli.commands.docs',
 'datateer_cli.commands.echo',
 'datateer_cli.commands.infra',
 'datateer_cli.commands.pipeline',
 'datateer_cli.orchestration']

package_data = \
{'': ['*'],
 'datateer_cli': ['ssh/*'],
 'datateer_cli.commands.pipeline': ['flow/*']}

install_requires = \
['PyYAML>=5.3.1,<6.0.0',
 'boto3>=1.24.46,<2.0.0',
 'click>=7.0',
 'google-cloud-storage>=2.5.0,<3.0.0',
 'pathspec>=0.9.0,<0.10.0']

extras_require = \
{'visualization': ['erd-python>=0.6.2,<0.7.0', 'pygraphviz>=1.10,<2.0']}

entry_points = \
{'console_scripts': ['datateer = datateer_cli.cli:main']}

setup_kwargs = {
    'name': 'datateer-cli',
    'version': '0.22.0',
    'description': 'Datateer CLI to support devops and infrastructure',
    'long_description': 'None',
    'author': 'Datateer',
    'author_email': 'dev@datateer.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import os
import subprocess
import tempfile

BACKEND_CONFIG_TEMPLATE = """
# This was automatically generated by running "datateer infra init-remote-state"
terraform {{
  backend "s3" {{
    bucket         = "datateer-terraform-remote-state-backend-{client_code}"
    dynamodb_table = "terraform-lock"
    encrypt        = true
    key            = "states/terraform.tfstate"
    profile        = "{aws_profile}"
    region         = "{aws_region}"
  }}
}}
"""


def configure_remote_state_backend(aws_profile, aws_region, client_code):
    if not os.path.exists("main.tf"):
        raise Exception(
            "Could not find file main.tf. Are you sure you are in the right directory? (<client-code>-infrastructure)"
        )
    command = ["terraform", "init", "-force-copy", "-backend=false"]
    subprocess.run(command)

    with tempfile.TemporaryDirectory() as tmp:
        backend_plan_path = os.path.join(tmp, "backend.plan")
        command = [
            "terraform",
            "plan",
            f"-out={backend_plan_path}",
            "-target=module.datateer_infrastructure.module.backend",
        ]
        subprocess.run(command)

        command = ["terraform", "apply", backend_plan_path]
        subprocess.run(command)

    f = open("main.tf", "r")
    backend_config_exists = 'backend "s3"' in f.read()
    f.close()
    if not backend_config_exists:
        backend_config = BACKEND_CONFIG_TEMPLATE.format(
            **{
                "aws_region": aws_region,
                "client_code": client_code,
                "aws_profile": aws_profile,
            }
        )
        with open("main.tf", "a") as f:
            f.write("")
            f.write(backend_config)
            f.write("")
    else:
        print("backend configuration already exists in main.tf, so not modifying it")

    command = ["terraform", "init", "-force-copy"]
    subprocess.run(command)



import click

from .operations import (
    aws_pull_config,
    aws_push_config,
    gcp_pull_config,
    gcp_push_config,
)

name = "Datateer CLI config commands"


@click.group(help="Commands related to configuration")
def config():
    pass


@config.command(
    help="Pulls configuration down from a Datateer environment and puts it into the configuration directory"
)
@click.option(
    "-d",
    "--config-dir",
    default=".datateer",
    help="will pull the config into this directory and overwrite what is there",
)
@click.option(
    "-r",
    "--region",
    envvar="REGION",
    help="(defaults to env var REGION)",
)
@click.option(
    "-e",
    "--environment",
    envvar="DATATEER_ENV",
    type=click.Choice(["prod", "stg", "qa", "int"], case_sensitive=False),
    required=True,
    help="The target Datateer environment (defaults to env var DATATEER_ENV)",
)
@click.option(
    "-b",
    "--config-bucket",
    help="Until the GCP implementation saves the config bucket name for us to dynamically retrieve it, we must rely on the caller to provide it",
)
@click.option(
    "-c",
    "--cloud",
    type=click.Choice(["aws", "gcp"], case_sensitive=False),
    required=True,
)
def pull(config_dir, region, environment, cloud, config_bucket):
    if cloud == "aws":
        aws_pull_config(config_dir, region, environment)
    elif cloud == "gcp":
        gcp_pull_config(config_dir, config_bucket, region)
    else:
        raise Exception(f'Value "{cloud}" was not recognized for argument "cloud"')


@config.command(
    help="Pushes configuration up to a Datateer environment from the configuration directory"
)
@click.option(
    "-d",
    "--config-dir",
    default=".datateer",
    help="will pull the config into this directory and overwrite what is there",
)
@click.option(
    "-r",
    "--region",
    envvar="REGION",
    help="(defaults to env var REGION)",
)
@click.option(
    "-e",
    "--environment",
    envvar="DATATEER_ENV",
    type=click.Choice(["prod", "stg", "qa", "int"], case_sensitive=False),
    required=True,
    help="The target Datateer environment (defaults to env var DATATEER_ENV)",
)
@click.option(
    "-b",
    "--config-bucket",
    help="Until the GCP implementation saves the config bucket name for us to dynamically retrieve it, we must rely on the caller to provide it",
)
@click.option(
    "-c",
    "--cloud",
    type=click.Choice(["aws", "gcp"], case_sensitive=False),
    required=True,
)
def push(config_dir, region, environment, cloud, config_bucket):
    if cloud == "aws":
        aws_push_config(config_dir, region, environment)
    elif cloud == "gcp":
        gcp_push_config(config_dir, config_bucket, region)
